arima = {}
catboost = {}
elasticnet = {}
gbt = {}
hwes = {}
knn = {}
lasso = {}
lightgbm = {}
lstm = {}
mlp = {}
mlr = {}
naive = {}
prophet = {}
rf = {}
rnn = {}
silverkite = {}
sgd = {}
svr= { }
tbats = {}
theta = {}
vecm = {}
xgboost = {}